package de.opalpiraten.scenepainter.sys;
import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.image.*;
import java.util.prefs.*;
import javax.imageio.*;
import java.beans.*;
import com.threerings.media.image.*;
import com.threerings.media.image.tools.xml.*;
import com.threerings.miso.util.*;
import com.threerings.miso.*;
import com.threerings.util.CompiledConfig;
import de.opalpiraten.scenepainter.gui.render.*;


/**
 * A Project is used to organize work by bundling scenes and supplementary
 * data together. It is also responsible for structuring the according files
 * on disk for persistent storage.
 */
public class Project implements ExceptionListener {

  /**
   * Name of the default Project
   */
  public static final String DEFAULTNAME="default";

  /**
   * name of this project
   */
  private String name;
  
  /**
   * The base path of this project.
   */
  private File bpath;
  
  /**
   * Pathname for saving scenes
   */
  public static final String SCENEDIR="scenes";
  
  /**
   * Pathname for saving sceneelements
   */
  public static final String SCENEELEMENTDIR="sceneelements";
  
  /**
   * Pathname for exporting PNG files to
   */
  public static final String EXPORTDIR="renderedscenes";
  
  /**
   * Pathname for imported image files
   */
  public static final String IMPORTDIR="importedgfx";
  
  /**
   * Pathname for additional nenya resources
   */
  public static final String NENYADIR="nenyarsrc";
  
  
  /**
   * Filename of the Colorpository in the NENYADIR
   */
  public static final String CPOSNAME="colordefs.xml";
  
  /**
   * Project specific colormappings
   */
  private ColorPository colorpository;
  
  /**
   * Create a new Project
   * @param name the name of the Project. If no Project by that name
   * exists, the according file structures will be created, otherwise they
   * are loaded.
   */
  public Project(String name) {
    this.name=name;
    bpath = new File(new File(new File(System.getProperty("user.dir")),"projects"),name);
    bpath.mkdirs();
    new File(bpath,SCENEDIR).mkdirs();
    new File(bpath,SCENEELEMENTDIR).mkdirs();
    new File(bpath,EXPORTDIR).mkdirs();
    new File(bpath,IMPORTDIR).mkdirs();
    new File(bpath,NENYADIR).mkdirs();
    
    
    File f = new File(new File(bpath,NENYADIR), CPOSNAME);
    if (f.isFile()) {
      try {
        //colorpository = ColorPository.loadColorPository(new FileInputStream(f));
        colorpository = (ColorPository) new ColorPositoryParser().parseConfig(f);
      }
      catch (Exception e) {
        e.printStackTrace();
      }
    }
  }
  
  /**
   * Query the project specific colorpository
   * @return either the colordefs.dat defined in the project or the
   * default ColorPository if none is set
   */
  public ColorPository getColorPository() {
    if (colorpository!=null)  return colorpository;
    else return Core.nenyabrowser.getColorPository();
  }
  
  /**
   * Assign a Colorpository to this Project. Calling this method will 
   * automatically dump the pository as an XML file in the projectfolder.
   */
  public void setColorPository(ColorPository c) {
    colorpository =c;
    
    // This should be done more elegantly
    StringBuffer sb = new StringBuffer();
    Iterator iter = colorpository.enumerateClasses();
    sb.append("<colors>\n");
    while (iter.hasNext()) {
      ColorPository.ClassRecord clazz= (ColorPository.ClassRecord)iter.next();
      sb.append("  <class classId=\""+clazz.classId+"\" name=\""+clazz.name+"\" source=\""+Util.toHexString(clazz.source)+"\" range=\""+clazz.range[0]+", "+clazz.range[1]+", "+clazz.range[2]+"\" starter=\""+clazz.starter+"\">\n");
      ColorPository.ColorRecord[] ccrec=colorpository.enumerateColors(clazz.name);
      for(int i=0;i<ccrec.length;i++) {
        sb.append("    <color colorId=\""+ccrec[i].colorId+"\" name=\""+ccrec[i].name+"\" offsets=\""+ccrec[i].offsets[0]+", "+ccrec[i].offsets[1]+", "+ccrec[i].offsets[2]+"\" starter=\""+ccrec[i].starter+"\"/>\n");
      }
      sb.append("  </class>\n");
    }
    sb.append("</colors>\n");
    
    try {
      File out = new File(new File(bpath,NENYADIR),CPOSNAME);
      FileOutputStream fos = new FileOutputStream(out);
      PrintStream ps = new PrintStream(fos);
      ps.print(sb.toString());
      File f = new File(new File(bpath,NENYADIR),"colordefs.dat");
      CompiledConfig.saveConfig(f, c);
      ps.flush();
      ps.close();
      colorpository = (ColorPository) new ColorPositoryParser().parseConfig(out);
    }
    catch (Exception e) {
      e.printStackTrace();
    }
  }
  
  /**
   * Query the name of this project
   */
  public String getName() {
    return name;
  }
  
  /**
   * Save a scene. This will also set the dirty flag of the scene to false;
   * @param s the scene to save (will be saved as XML).
   * @param e export if true, the scene is also exported as PNG
   */
  public void saveScene(Scene s, boolean e) throws IOException {
    File f = getFileFor(SCENEDIR,s.getName());
    
    XMLEncoder enc = new XMLEncoder(new BufferedOutputStream(new FileOutputStream(f)));
    enc.writeObject(Scene.strip(s));
    enc.close();
    s.dirty=false;    
    if(e) {
      f = getFileFor(EXPORTDIR,s.getName());
      ImageIO.write(s.export(),"png",f);
    }
  }
  
  /**
   * Load a scene
   * @param name Filename of the scene to load (without XML suffix).
   */
  public Scene loadScene(String name) 
  throws FileNotFoundException, IOException {
    File f = getFileFor(SCENEDIR,name);
    XMLDecoder dec = new XMLDecoder(new BufferedInputStream(new FileInputStream(f)));
    dec.setExceptionListener(this);
    Scene ret = (Scene)dec.readObject();
    // The name of the scene must equal the filename. if it does not, the
    // filename is authorative (the user probably copied it, hoping to rename
    // it - honor his/her wish).
    ret.setName(name);
    dec.close();
    SceneElement[] elem = ret.getElements();
    for(int i=0;i<elem.length;i++) elem[i].setProject(this);
    ret.dirty=false;
    return ret;
  }
  
  /**
   * Delete a scene
   * @param name of the scene to delete
   */
  public void deleteScene(String name) 
  throws FileNotFoundException, IOException {
    File f = new File(new File(bpath,SCENEDIR),name+".xml");
    f.delete();
  }
  
  /**
   * Load a SceneElement.
   * @param name name of the file to load (wihtout .XML suffix).
   */
  public SceneElement loadSceneElement(String name)
  throws FileNotFoundException, IOException {
    File f = getFileFor(SCENEELEMENTDIR,name);
    XMLDecoder dec = new XMLDecoder(new BufferedInputStream(new FileInputStream(f)));
    dec.setExceptionListener(this);
    SceneElement ret = (SceneElement)dec.readObject();
    dec.close();
    // The name of the scene must equal the filename. if it does not, the
    // filename is authorative (the user probably copied it, hoping to rename
    // it - honor his/her wish).
    ret.setName(name);
    ret.setProject(this);
    ret.setPosition(new Point(0,0));
    return ret;
  }
  
  /**
   * Save a SceneElement
   */
  public void saveSceneElement(SceneElement s) throws IOException {
    File f = getFileFor(SCENEELEMENTDIR,s.getName());
    XMLEncoder enc = new XMLEncoder(new BufferedOutputStream(new FileOutputStream(f)));
    enc.writeObject(s);
    enc.close();
  }
  
  /**
   * Delete a sceneelement
   * @param name of the sceneelement to delete
   */
  public void deleteSceneElement(String name) 
  throws FileNotFoundException, IOException {
    File f = getFileFor(SCENEELEMENTDIR,name);
    f.delete();
  }
  
  /**
   * Import PNG file(s) to this Projekt.
   */
  public void importPicture(File[] f) throws IOException, FileNotFoundException {
    for (int c=0;c<f.length;c++) {
      File out = new File(new File(bpath,IMPORTDIR),f[c].getName());
      out.createNewFile();
      byte[] buf = new byte[16384];
      FileInputStream fis  = new FileInputStream(f[c]);
      FileOutputStream fos = new FileOutputStream(out);
      int i=0;
      while((i=fis.read(buf))!=-1) {
        fos.write(buf, 0, i);
      }
      fis.close();
      fos.close();
    }
  }
  
  
  /**
   * Load a picture.
   * @param name the name of the file, the picture is stored in 
   * (wihtout .png extension).
   */
  public BufferedImage loadPicture(String name) throws IOException {
    File f = new File(new File(bpath,IMPORTDIR),name+".png");
    return ImageIO.read(f);
  }
  
  /**
   * List the names of the Scenes stored in this project
   * @return the names of all available scenes.
   */
  public String[] listScenes() {
    File f = new File(bpath,SCENEDIR);
    File[] tmp = f.listFiles(new SuffixFilter(SuffixFilter.XML,false));
    String ret[] = new String[tmp.length];
    for (int i=0;i<tmp.length;i++) {
      ret[i] = tmp[i].getName().substring(0,tmp[i].getName().length()-4);
    }
    Arrays.sort(ret);
    return ret;
  }
  
  /**
   * List the names of the SceneElements stored in this Project
   * @return the filenames of all available elements.
   */
  public String[] listSceneElements() {
    File f;
    Preferences cfg = Preferences.userNodeForPackage(Project.class);
    f = new File(cfg.get("centralelementdir",new File(bpath,SCENEELEMENTDIR).getAbsolutePath()));
    /*
    if (centralelementdir==null) {
      f= new File(bpath,SCENEELEMENTDIR);
    }
    else {
      f = centralelementdir;
    }*/
    File[] tmp = f.listFiles(new SuffixFilter(SuffixFilter.XML,false));
    String ret[] = new String[tmp.length];
    for (int i=0;i<tmp.length;i++) {
      ret[i] = tmp[i].getName().substring(0,tmp[i].getName().length()-4);
    }
    Arrays.sort(ret);
    return ret;
  }
  
  /**
   * List all pictures (PNG files) imported into this Project
   * @return a list of the filenames without path and suffix, that can be
   * loaded via loadPicture().
   */
  public String[] listPictures() {
    File f = new File(bpath,IMPORTDIR);
    File[] tmp = f.listFiles(new SuffixFilter(SuffixFilter.PNG,false));
    String ret[] = new String[tmp.length];
    for (int i=0;i<tmp.length;i++) {
      ret[i] = tmp[i].getName().substring(0,tmp[i].getName().length()-4);
    }
    Arrays.sort(ret);
    return ret;
  }
  
  /**
   * List all Projects in the project folder
   */
  public static String[] listProjects() {
    String[] tmp= new Project(DEFAULTNAME).bpath.getParentFile().list();
    Arrays.sort(tmp);
    return tmp;
  }
  
  // Interface implementation
  public void exceptionThrown(Exception e) {
    if (e instanceof NoSuchMethodException) {
      //System.err.println("Missmatch in XML file, resaving will fix it:\n"+e.getMessage());
    }
    else e.printStackTrace();
  }
  
  
  /**
   * Query the storage location, this project would or does assign to a
   * certain object.
   * @param type Type of the object (e.g. Project.SCENEDIR). Requesting a 
   * bogus type will make this method return the base directory of this
   * Project.
   * @param name Name of the object, the storage location is requested for.
   * If null is submitted instead of a name, the returned location will be
   * the storing directory
   * @return a pointer to the file of the requested object. Note, that this
   * is just a pointer, the actual file may or may not exist (yet).
   */
  public File getFileFor(String type, String name) {
    File ret = bpath;
    
    if (type==null) return ret;
    
    Preferences cfg = Preferences.userNodeForPackage(Project.class);
    
    
    if (type.equals(SCENEDIR)) {
      File tmp = new File(bpath,SCENEDIR);
      if (name!=null) ret = new File(tmp,name+".xml");
      else ret = tmp;
    }
    
    if (type.equals(SCENEELEMENTDIR)) {
      File tmp = new File(cfg.get("centralelementdir",new File(bpath,SCENEELEMENTDIR).getAbsolutePath()));
      if (name!=null) ret = new File(tmp,name+".xml");
      else ret = tmp; 
    }
    
    if (type.equals(EXPORTDIR)) {
      File tmp = new File(cfg.get("centralexportdir",new File(bpath,EXPORTDIR).getAbsolutePath()));
      if (name!=null) ret = new File(tmp,name+".png");
      else ret = tmp; 
    }
    
    if (type.equals(IMPORTDIR)) {
      File tmp = new File(bpath,IMPORTDIR);
      if (name!=null) ret = new File(tmp,name+".png");
      else ret = tmp;
    }
    
    if (type.equals(NENYADIR)) {
      File tmp = new File(bpath,NENYADIR);
      if (name!=null) ret = new File(tmp,name);
      else ret = tmp;
    }
    
    return ret;
  }
}
